import re


def value_from_stdout(field, stdout):
    """Extract the output from the stdout using the field name and the output directory
    to be used as a callable in output specs

    Parameters
    ----------
    field : attrs.field
        the output field
    stdout : str
        the stdout from the container

    Returns
    -------
    value : Any
        the value to be used as the output
    """
    try:
        regex = field.metadata["value_regex"]
    except KeyError:
        value = stdout
    else:
        value = re.match(regex, stdout)
    if field.type:
        value = field.type(value)
    return value



from __future__ import annotations
import typing as ty
from pathlib import Path
import attrs
from pydra import ShellCommandTask
from pydra.engine.specs import SpecInfo, ShellSpec, ShellOutSpec
from frametree.core.serialize import ClassResolver, ObjectListConverter
from fileformats.core import FileSet, DataType, Field
from pydra2app.core import PACKAGE_NAME


@attrs.define(kw_only=True)
class ShellCmdField:
    name: str
    datatype: type = attrs.field(
        converter=ClassResolver(DataType, alternative_types=[bool, str, float, int])
    )
    help_string: str = ""

    def attrs_metadata(self, skip_fields=None):
        if skip_fields is None:
            skip_fields = []
        skip_fields += ["name", "datatype"]
        metadata = {
            n: v
            for n, v in attrs.asdict(self).items()
            if n not in skip_fields and v is not None
        }
        if issubclass(self.datatype, FileSet) and "argstr" not in metadata:
            metadata["argstr"] = ""
        return metadata

    @property
    def attrs_type(self):
        if issubclass(self.datatype, FileSet):
            tp = self.datatype
        elif self.datatype.__module__ == "builtins":
            tp = self.datatype
        elif issubclass(self.datatype, Field):
            tp = self.datatype.primitive
        else:
            raise ValueError(f"Unsupported shell command input type {self.datatype}")
        return tp


@attrs.define(kw_only=True)
class ShellCmdInput(ShellCmdField):
    """Specifies an input field to the shell command

    Parameters
    ----------
    name : str
        the name of the input field
    datatype : type
        the type of the input field
    mandatory (bool, default: False):
        If True user has to provide a value for the field.
    sep (str):
        A separator if a list is provided as a value.
    argstr (str):
        A flag or string that is used in the command before the value, e.g. -v or
        -v {inp_field}, but it could be and empty string, “”. If … are used, e.g. -v…,
        the flag is used before every element if a list is provided as a value. If no
        argstr is used the field is not part of the command.
    position (int):
        Position of the field in the command, could be nonnegative or negative integer.
        If nothing is provided the field will be inserted between all fields with
        nonnegative positions and fields with negative positions.
    allowed_values (list):
        List of allowed values for the field.
    requires (list):
        List of field names that are required together with the field.
    xor (list):
        List of field names that are mutually exclusive with the field.
    copyfile (bool, default: False):
        If True, a hard link is created for the input file in the output directory. If
        hard link not possible, the file is copied to the output directory.
    container_path (bool, default: False, only for ContainerTask):
        If True a path will be consider as a path inside the container (and not as a
        local path).
    output_file_template (str):
        If provided, the field is treated also as an output field and it is added to the
        output spec. The template can use other fields, e.g. {file1}. Used in order to
        create an output specification.
    output_field_name (str, used together with output_file_template)
        If provided the field is added to the output spec with changed name. Used in
        order to create an output specification.
    keep_extension (bool, default: True):
        A flag that specifies if the file extension should be removed from the field
        value. Used in order to create an output specification.
    readonly (bool, default: False):
        If True the input field can’t be provided by the user but it aggregates other
        input fields (for example the fields with argstr: -o {fldA} {fldB}).
    formatter (function)
        If provided the argstr of the field is created using the function. This function
        can for example be used to combine several inputs into one command argument.
        The function can take field (this input field will be passed to the function),
        inputs (entire inputs will be passed) or any input field name (a specific input
        field will be sent).
    """

    sep: ty.Optional[str] = None
    argstr: ty.Optional[str] = None
    position: ty.Optional[int] = None
    allowed_values: ty.Optional[list] = None
    requires: ty.Optional[list] = None
    xor: ty.Optional[list] = None
    copyfile: ty.Optional[bool] = None
    mandatory: ty.Optional[bool] = None
    readonly: ty.Optional[bool] = None
    formatter: ty.Callable = attrs.field(
        default=None,
        converter=ClassResolver(allow_none=True, package=PACKAGE_NAME),
    )


@attrs.define(kw_only=True)
class ShellCmdOutput(ShellCmdField):
    """Specifies an output field from the shell command

    Parameters
    ----------
    name : str
        the name of the input field
    datatype : type
        the type of the input field
    mandatory : bool, default: False
        If True the output file has to exist, otherwise an error will be raised.
    output_file_template : str
        If provided the output file name (or list of file names) is created using the
        template. The template can use other fields, e.g. {file1}. The same as in input_spec.
    output_field_name : str
        If provided the field is added to the output spec with changed name. The same as in
        input_spec.
    keep_extension : bool, default: True
        A flag that specifies if the file extension should be removed from the field value.
        The same as in input_spec.
    requires : list
        List of field names that are required to create a specific output. The fields do not
        have to be a part of the output_file_template and if any field from the list is not
        provided in the input, a NOTHING is returned for the specific output. This has a
        different meaning than the requires form the input_spec.
    callable : function
        If provided the output file name (or list of file names) is created using the
        function. The function can take field (the specific output field will be passed
        to the function), output_dir (task output_dir will be used), stdout, stderr
        (stdout and stderr of the task will be sent) inputs (entire inputs will be
        passed) or any input field name (a specific input field will be sent).
    """

    argstr: bool = attrs.field(default=None, metadata={"input": True})
    mandatory: ty.Optional[bool] = None
    position: int = attrs.field(default=None, metadata={"input": True})
    output_file_template: ty.Optional[str] = attrs.field()
    output_field_name: ty.Optional[str] = None
    keep_extension: bool = attrs.field(default=None, metadata={"input": True})
    requires: ty.Optional[list] = None
    formatter: ty.Callable = attrs.field(
        default=None,
        converter=ClassResolver(allow_none=True, package=PACKAGE_NAME),
        metadata={"input": True},
    )
    callable: ty.Callable = attrs.field(
        default=None, converter=ClassResolver(allow_none=True, package=PACKAGE_NAME)
    )

    @output_file_template.default
    def output_file_template_default(self):
        if issubclass(self.datatype, FileSet):
            return self.name + self.datatype.strext
        else:
            return None

    @property
    def input_required(self):
        input_only_values = [getattr(self, f) for f in self.input_only_fields]
        return any(v is not None for v in input_only_values)

    @property
    def input_only_fields(self):
        return [f.name for f in attrs.fields(type(self)) if f.metadata.get("input")]

    # @property
    # def attrs_type(self):
    #     tp = super().attrs_type
    #     # FIXME: this shouldn't be necessary. Pydra should be smart enough that this is
    #     # an output file and shouldn't need to exist
    #     if tp in (pydra.engine.specs.File, pydra.engine.specs.Directory):
    #         tp = str
    #     return tp


def shell(
    name: str,
    inputs: ty.List[ty.Union[ShellCmdInput, ty.Dict[str, str]]],
    outputs: ty.List[ty.Union[ShellCmdOutput, ty.Dict[str, str]]],
    executable: str = "",  # Use entrypoint of container,
    parameters: ty.List[ty.Union[ShellCmdInput, ty.Dict[str, str]]] = None,
    **kwargs,
):
    """Creates a Pydra shell command task which takes file inputs and runs it on the
    provided inputs, outputs and parameters

    Parameters
    ----------
    name : str
        Name of the workflow/BIDS app. Will be used to name the 'derivatives'
        sub-directory where the app outputs are stored
    inputs : list[tuple[str, type] or ty.Dict[str, str]]
        The inputs to be inserted into the BIDS dataset. Should be a list of tuples
        consisting of the the path the file/directory should be stored within a BIDS subject/session,
        e.g. anat/T1w, func/bold, and the DataFormat class it should be stored in, e.g.
        file.NiftiGzX.
    outputs : list[tuple[str, type] or ty.Dict[str, str]]
        The outputs to be extracted from the derivatives directory. Should be a list of tuples
        consisting of the the path the file/directory is saved by the app within a BIDS subject/session,
        e.g. freesurfer/recon-all, and the DataFormat class it is stored in
    executable : str, optional
        Name of the executable within the image to run (i.e. the entrypoint of the image).
        Required when extending the base image and launching Arcana within it. Defaults to
        empty string, i.e. the entrypoint of the BIDS app container image
    parameters : dict[str, type], optional
        a list of parameters of the app (i.e. CLI flags) to be exposed to the user
        mapped to their data type.
    row_frequency : Clinical, optional
        Frequency to run the app at, i.e. per-"session" or per-"dataset"

    Returns
    -------
    pydra.ShellCommmandTask
        A Pydra shell command task that can be deployed using the deployment framework
    """
    inputs = ObjectListConverter(ShellCmdInput)(inputs)
    outputs = ObjectListConverter(ShellCmdOutput)(outputs)
    parameters = ObjectListConverter(ShellCmdInput, allow_none=True)(parameters)

    input_fields = []
    for inpt in inputs + parameters:
        input_fields.append(
            (
                inpt.name,
                inpt.attrs_type,
                inpt.attrs_metadata(),
            )
        )

    output_fields = []
    for outpt in outputs:
        if outpt.input_required:
            if issubclass(outpt.datatype, FileSet):
                tp = Path
            else:
                tp = outpt.datatype
            input_fields.append(
                (
                    outpt.name,
                    tp,
                    outpt.attrs_metadata(),
                )
            )
        output_fields.append(
            (
                outpt.name,
                outpt.attrs_type,
                outpt.attrs_metadata(skip_fields=outpt.input_only_fields),
            )
        )

    return ShellCommandTask(
        name=name,
        executable=executable,
        input_spec=SpecInfo(name="Input", fields=input_fields, bases=(ShellSpec,)),
        output_spec=SpecInfo(
            name="Output", fields=output_fields, bases=(ShellOutSpec,)
        ),
        **kwargs,
    )



import os
import docker
from frametree.common import DirTree, Samples
from pydra2app.core.image import App
from pydra2app.core import __version__, PACKAGE_NAME


def test_native_python_install(tmp_path):

    SAMPLE_INDEX = "1"
    OUTPUT_COL_NAME = "printed_version"

    dataset_dir = tmp_path / "dataset"
    sample_dir = dataset_dir / SAMPLE_INDEX
    sample_dir.mkdir(parents=True)
    sample_file = sample_dir / "sample.txt"
    sample_file.write_text("sample")

    dataset = DirTree().define_dataset(dataset_dir, space=Samples)
    dataset.save()

    test_spec = {
        "name": "native_python_test",
        "title": "a test image spec",
        "command": {
            "task": "common:shell",
            "inputs": {
                "dummy": {
                    "datatype": "text/text-file",
                    "help": "a dummy input that isn't actually used",
                    "configuration": {
                        "position": 0,
                    },
                },
            },
            "outputs": {
                OUTPUT_COL_NAME: {
                    "datatype": "field/text",
                    "help": "the print to stdout",
                    "configuration": {
                        "callable": "common:value_from_stdout",
                    },
                }
            },
            "parameters": {
                "duplicates": {
                    "field": "duplicates",
                    "default": 2,
                    "datatype": "field/integer",
                    "required": True,
                    "help": "a parameter",
                }
            },
            "row_frequency": "common:Samples[sample]",
            "configuration": {
                "executable": [
                    "pydra2app",
                    "--version",
                ]
            },
        },
        "version": {"package": "1.0", "build": "1"},
        "packages": {
            "system": ["vim"],  # just to test it out
            "pip": {"pydra2app": None, "frametree": None},  # just to test out the
        },
        "base_image": {
            "name": "python",
            "tag": "3.12.5-slim-bookworm",
            "python": "python3",
            "package_manager": "apt",
            "conda_env": None,
        },
        "authors": [{"name": "Some One", "email": "some.one@an.email.org"}],
        "docs": {
            "info_url": "http://concatenate.readthefakedocs.io",
        },
    }

    app = App.load(test_spec)

    app.make(build_dir=tmp_path / "build-dir", use_local_packages=True)

    volume_mount = str(dataset_dir) + ":/dataset:rw"
    args = ["/dataset", "--input", "dummy", "sample"]

    dc = docker.from_env()
    try:
        dc.containers.run(
            app.reference,
            command=args,
            stderr=True,
            volumes=[volume_mount],
            user=f"{os.getuid()}:{os.getgid()}",
        )
    except docker.errors.ContainerError as e:
        raise RuntimeError(
            f"'docker run -v {volume_mount} {app.reference} {' '.join(args)}' errored:\n"
            + e.stderr.decode("utf-8")
        )

    dataset = DirTree().load_dataset(dataset_dir)

    def strip_ver_timestamp(ver_str):
        parts = str(ver_str).split("+")
        try:
            parts[1] = parts[1].split(".")[0]
        except IndexError:
            pass
        return "+".join(parts).strip()

    assert (
        strip_ver_timestamp(dataset[OUTPUT_COL_NAME][SAMPLE_INDEX])
        == f"{PACKAGE_NAME}, version {strip_ver_timestamp(__version__)}"
    )



from ._version import __version__

PACKAGE_NAME = "pydra2app"
CODE_URL = f"https://github.com/ArcanaFramework/{PACKAGE_NAME}"


__authors__ = [("Thomas G. Close", "tom.g.close@gmail.com")]

from .image import Pydra2AppImage, App, Metapackage  # noqa



import sys
import logging
import shutil
from pathlib import Path
import json
import typing as ty
import re
from collections import defaultdict
from traceback import format_exc
import tempfile
import click
import yaml
import docker
import docker.errors
from pydra.engine.core import TaskBase
from frametree.core.serialize import (
    package_from_module,
    pydra_asdict,
    ClassResolver,
)
from frametree.core.packaging import submodules
import pydra2app
from pydra2app.core import __version__
from pydra2app.core.image import Metapackage, App
from pydra2app.core.exceptions import Pydra2AppBuildError
from pydra2app.core.utils import extract_file_from_docker_image, DOCKER_HUB
from pydra2app.core.command import entrypoint_opts
from pydra2app.core import PACKAGE_NAME


logger = logging.getLogger("pydra2app")


# Define the base CLI entrypoint
@click.group()
@click.version_option(version=__version__)
def cli():
    """Base command line group, installed as "pydra2app"."""


@cli.command(
    name="make",
    help="""Construct and build a docker image containing a pipeline to be run on data
stored in a data repository or structure (e.g. XNAT Container Service Pipeline or BIDS App)

TARGET is the type of image to build. For standard images just the pydra2app
sub-package is required (e.g. 'xnat' or 'common'). However, specific App subclasses can
be specified using <module-path>:<app-class-name> format, e.g. pydra2app.xnat:XnatApp

SPEC_PATH is the file system path to the specification to build, or directory
containing multiple specifications
""",
)
@click.argument("target", type=str)
@click.argument("spec_path", type=click.Path(exists=True, path_type=Path))
@click.option(
    "--registry",
    default=DOCKER_HUB,
    help="The Docker registry to deploy the pipeline to",
)
@click.option(
    "--build-dir",
    default=None,
    type=click.Path(path_type=Path),
    help=(
        "Specify the directory to build the Docker image in. "
        "Defaults to `.build` in the directory containing the "
        "YAML specification"
    ),
)
@click.option(
    "--release",
    default=None,
    nargs=2,
    metavar="<release-name> <release-version>",
    type=str,
    help=("Name of the release for the package as a whole (i.e. for all pipelines)"),
)
@click.option(
    "--tag-latest/--dont-tag-latest",
    default=False,
    type=bool,
    help='whether to tag the release as the "latest" or not',
)
@click.option(
    "--save-manifest",
    default=None,
    type=click.Path(writable=True),
    help="File path at which to save the build manifest",
)
@click.option(
    "--logfile",
    default=None,
    type=click.Path(path_type=Path),
    help="Log output to file instead of stdout",
)
@click.option("--loglevel", default="info", help="The level to display logs at")
@click.option(
    "--use-local-packages/--dont-use-local-packages",
    type=bool,
    default=False,
    help=(
        "Use locally installed Python packages, instead of pulling "
        "them down from PyPI"
    ),
)
@click.option(
    "--install-extras",
    type=str,
    default=None,
    help=(
        "Install extras to use when installing Pydra2App inside the "
        "container image. Typically only used in tests to provide "
        "'test' extra"
    ),
)
@click.option(
    "--for-localhost/--not-for-localhost",
    type=bool,  # FIXME: This should be replaced with option to set XNAT CS IP address
    default=False,
    help=(
        "Build the image so that it can be run in Pydra2App's test "
        "configuration (only for internal use)"
    ),
)
@click.option(
    "--raise-errors/--log-errors",
    type=bool,
    default=False,
    help=("Raise exceptions instead of logging failures"),
)
@click.option(
    "--generate-only/--build",
    type=bool,
    default=False,
    help="Just create the build directory and dockerfile",
)
@click.option(
    "--license",
    type=(str, click.Path(exists=True, path_type=Path)),
    default=(),
    nargs=2,
    metavar="<license-name> <path-to-license-file>",
    multiple=True,
    help=(
        "Licenses provided at build time to be stored in the image (instead of "
        "downloaded at runtime)"
    ),
)
@click.option(
    "--license-to-download",
    type=str,
    default=(),
    multiple=True,
    help=(
        "Specify licenses that are not provided at runtime and instead downloaded "
        "from the data store at runtime in order to satisfy their conditions"
    ),
)
@click.option(
    "--check-registry/--dont-check-registry",
    type=bool,
    default=False,
    help=(
        "Check the registry to see if an existing image with the "
        "same tag is present, and if so whether the specification "
        "matches (and can be skipped) or not (raise an error)"
    ),
)
@click.option(
    "--push/--dont-push",
    type=bool,
    default=False,
    help=("push built images to registry"),
)
@click.option(
    "--clean-up/--dont-clean-up",
    type=bool,
    default=False,
    help=("Remove built images after they are pushed to the registry"),
)
@click.option(
    "--spec-root",
    type=click.Path(path_type=Path, exists=True),
    default=None,
    help=("The root path to consider the specs to be relative to, defaults to CWD"),
)
@click.option(
    "--source-package",
    "-s",
    type=click.Path(path_type=Path, exists=True),
    multiple=True,
    default=(),
    help=(
        "Path to a local Python package to be included in the image. Needs to have a "
        "package definition that can be built into a source distribution and the name of "
        "the directory needs to match that of the package to be installed. Multiple "
        "packages can be specified by repeating the option."
    ),
)
@click.option(
    "--export-file",
    "-e",
    "export_files",
    type=str,
    nargs=2,
    multiple=True,
    default=(),
    metavar="<internal-dir> <external-dir>",
    help=(
        "Path to be exported from the Docker build directory for convenience. Multiple "
        "files can be specified by repeating the option."
    ),
)
def make(
    target,
    spec_path: Path,
    registry,
    release,
    tag_latest,
    save_manifest,
    logfile,
    loglevel,
    build_dir: Path,
    use_local_packages,
    install_extras,
    raise_errors,
    generate_only,
    for_localhost,
    license,
    license_to_download,
    check_registry,
    push,
    clean_up,
    spec_root: Path,
    source_package: ty.Sequence[Path],
    export_files: ty.Sequence[ty.Tuple[Path, Path]],
):

    if isinstance(spec_path, bytes):  # FIXME: This shouldn't be necessary
        spec_path = Path(spec_path.decode("utf-8"))
    if isinstance(build_dir, bytes):  # FIXME: This shouldn't be necessary
        build_dir = Path(build_dir.decode("utf-8"))

    if tag_latest and not release:
        raise ValueError("'--tag-latest' flag requires '--release'")

    if spec_root is None:
        if spec_path.is_file():
            spec_root = spec_path.parent.parent
        else:
            spec_root = spec_path.parent
        logger.info(
            "`--spec-root` was not explicitly provided so assuming it is the parent '%s'",
            str(spec_root),
        )

    path_parts = spec_path.relative_to(spec_root).parts

    if spec_path.is_file() and len(path_parts) < 2:
        raise ValueError(
            f"Spec paths ({spec_path}) must be placed within (a) nested director(y|ies) "
            "from the spec root. The top-level nested directory will be interpreted as "
            "the name of the Docker package and subsequent directories will be used to "
            "qualify the image name with '.' separated prefixes"
        )

    package_name = path_parts[0]

    if build_dir is None:
        if spec_path.is_file():
            build_dir = spec_path.parent / (".build-" + spec_path.stem)
        else:
            build_dir = spec_path / ".build"

    if not build_dir.exists():
        build_dir.mkdir()

    install_extras = install_extras.split(",") if install_extras else []

    logging.basicConfig(filename=logfile, level=getattr(logging, loglevel.upper()))

    temp_dir = tempfile.mkdtemp()

    target_cls: App = ClassResolver(App, package=PACKAGE_NAME)(target)

    dc = docker.from_env()

    license_paths = {}
    for lic_name, lic_src in license:
        if isinstance(lic_src, bytes):  # FIXME: This shouldn't be necessary
            lic_src = Path(lic_src.decode("utf-8"))
        license_paths[lic_name] = lic_src

    # Load image specifications from YAML files stored in directory tree

    # Don't error if the modules the task, data stores, data types, etc...
    # aren't present in the build environment
    # FIXME: need to test for this
    with ClassResolver.FALLBACK_TO_STR:
        image_specs: ty.List[target_cls] = target_cls.load_tree(
            spec_path,
            root_dir=spec_root,
            registry=registry,
            license_paths=license_paths,
            licenses_to_download=set(license_to_download),
            source_packages=source_package,
        )

    # Check the target registry to see a) if the images with the same tag
    # already exists and b) whether it was built with the same specs
    if check_registry:
        conflicting = {}
        to_build = []
        for image_spec in image_specs:
            try:
                extracted_file = extract_file_from_docker_image(
                    image_spec.reference, image_spec.IN_DOCKER_SPEC_PATH
                )
            except docker.errors.NotFound:
                extracted_file = None
            if extracted_file is None:
                logger.info(
                    f"Did not find existing image matching {image_spec.reference}"
                )
                changelog = None
            else:
                logger.info(
                    f"Comparing build spec with that of existing image {image_spec.reference}"
                )
                built_spec = image_spec.load(extracted_file)

                changelog = image_spec.compare_specs(built_spec, check_version=True)

            if changelog is None:
                to_build.append(image_spec)
            elif not changelog:
                logger.info(
                    "Skipping '%s' build as identical image already "
                    "exists in registry"
                )
            else:
                conflicting[image_spec.reference] = changelog

        if conflicting:
            msg = ""
            for tag, changelog in conflicting.items():
                msg += (
                    f"spec for '{tag}' doesn't match the one that was "
                    "used to build the image already in the registry:\n\n"
                    + str(changelog.pretty())
                    + "\n\n\n"
                )

            raise Pydra2AppBuildError(msg)

        image_specs = to_build

    if release or save_manifest:
        manifest = {
            "package": package_name,
            "images": [],
        }
        if release:
            manifest["release"] = ":".join(release)

    errors = False

    for image_spec in image_specs:
        spec_build_dir = (
            build_dir / image_spec.loaded_from.relative_to(spec_path.absolute())
        ).with_suffix("")
        if spec_build_dir.exists():
            shutil.rmtree(spec_build_dir)
        spec_build_dir.mkdir(parents=True)
        try:
            image_spec.make(
                build_dir=spec_build_dir,
                for_localhost=for_localhost,
                use_local_packages=use_local_packages,
                generate_only=generate_only,
                no_cache=clean_up,
            )
        except Exception:
            if raise_errors:
                raise
            logger.error(
                "Could not build %s pipeline:\n%s", image_spec.reference, format_exc()
            )
            errors = True
            continue
        else:
            click.echo(image_spec.reference)
            logger.info("Successfully built %s pipeline", image_spec.reference)

        if push:
            try:
                dc.api.push(image_spec.reference)
            except Exception:
                if raise_errors:
                    raise
                logger.error(
                    "Could not push '%s':\n\n%s", image_spec.reference, format_exc()
                )
                errors = True
            else:
                logger.info(
                    "Successfully pushed '%s' to registry", image_spec.reference
                )
        if clean_up:

            def remove_image_and_containers(image_ref):
                logger.info(
                    "Removing '%s' image and associated containers to free up disk space "
                    "as '--clean-up' is set",
                    image_ref,
                )
                for container in dc.containers.list(filters={"ancestor": image_ref}):
                    container.stop()
                    container.remove()
                dc.images.remove(image_ref, force=True)
                result = dc.containers.prune()
                dc.images.prune(filters={"dangling": False})
                logger.info(
                    "Removed '%s' image and associated containers and freed up %s of disk space ",
                    image_ref,
                    result["SpaceReclaimed"],
                )

            remove_image_and_containers(image_spec.reference)
            remove_image_and_containers(image_spec.base_image.reference)

        if release or save_manifest:
            manifest["images"].append(
                {
                    "name": image_spec.path,
                    "version": image_spec.tag,
                }
            )
    if release:
        metapkg = Metapackage(
            name=release[0],
            version=release[1],
            org=package_name,
            manifest=manifest,
        )
        metapkg.make(use_local_packages=use_local_packages)
        if push:
            try:
                dc.api.push(metapkg.reference)
            except Exception:
                if raise_errors:
                    raise
                logger.error(
                    "Could not push release metapackage '%s':\n\n%s",
                    metapkg.reference,
                    format_exc(),
                )
                errors = True
            else:
                logger.info(
                    "Successfully pushed release metapackage '%s' to registry",
                    metapkg.reference,
                )

            if tag_latest:
                # Also push release to "latest" tag
                image = dc.images.get(metapkg.reference)
                latest_tag = metapkg.path + ":latest"
                image.tag(latest_tag)

                try:
                    dc.api.push(latest_tag)
                except Exception:
                    if raise_errors:
                        raise
                    logger.error(
                        "Could not push latest tag for release metapackage '%s':\n\n%s",
                        metapkg.path,
                        format_exc(),
                    )
                    errors = True
                else:
                    logger.info(
                        (
                            "Successfully pushed latest tag for release metapackage '%s' "
                            "to registry"
                        ),
                        metapkg.path,
                    )
        if save_manifest:
            with open(save_manifest, "w") as f:
                json.dump(manifest, f, indent="    ")

    for src_path, dest_path in export_files:
        dest_path = Path(dest_path)
        dest_path.parent.mkdir(parents=True, exist_ok=True)
        full_src_path = build_dir / src_path
        if not full_src_path.exists():
            logger.warning(
                "Could not find file '%s' to export from build directory", full_src_path
            )
            continue
        if full_src_path.is_dir():
            shutil.copytree(full_src_path, dest_path)
        else:
            shutil.copy(full_src_path, dest_path)

    shutil.rmtree(temp_dir)
    if errors:
        sys.exit(1)


@cli.command(
    name="list-images",
    help="""Walk through the specification paths and list tags of the images
that will be build from them.

SPEC_ROOT is the file system path to the specification to build, or directory
containing multiple specifications

DOCKER_ORG is the Docker organisation the images should belong to""",
)
@click.argument("spec_root", type=click.Path(exists=True, path_type=Path))
@click.option(
    "--registry",
    default=None,
    help="The Docker registry to deploy the pipeline to",
)
def list_images(spec_root, registry):
    if isinstance(spec_root, bytes):  # FIXME: This shouldn't be necessary
        spec_root = Path(spec_root.decode("utf-8"))

    for image_spec in App.load_tree(spec_root, registry=registry):
        click.echo(image_spec.reference)


@cli.command(
    name="make-docs",
    help="""Build docs for one or more yaml wrappers

SPEC_ROOT is the path of a YAML spec file or directory containing one or more such files.

The generated documentation will be saved to OUTPUT.
""",
)
@click.argument("spec_path", type=click.Path(exists=True, path_type=Path))
@click.argument("output", type=click.Path(path_type=Path))
@click.option(
    "--registry",
    default=DOCKER_HUB,
    help="The Docker registry to deploy the pipeline to",
)
@click.option("--flatten/--no-flatten", default=False)
@click.option("--loglevel", default="warning", help="The level to display logs at")
@click.option(
    "--default-data-space",
    default=None,
    help=(
        "The default data space to assume if it isn't explicitly stated in "
        "the command"
    ),
)
@click.option(
    "--spec-root",
    type=click.Path(path_type=Path),
    default=None,
    help=("The root path to consider the specs to be relative to, defaults to CWD"),
)
def make_docs(
    spec_path, output, registry, flatten, loglevel, default_data_space, spec_root
):
    # # FIXME: Workaround for click 7.x, which improperly handles path_type
    # if type(spec_path) is bytes:
    #     spec_path = Path(spec_path.decode("utf-8"))
    # if type(output) is bytes:
    #     output = Path(output.decode("utf-8"))

    logging.basicConfig(level=getattr(logging, loglevel.upper()))

    output.mkdir(parents=True, exist_ok=True)

    default_data_space = ClassResolver.fromstr(default_data_space)

    with ClassResolver.FALLBACK_TO_STR:
        image_specs = App.load_tree(
            spec_path,
            registry=registry,
            root_dir=spec_root,
            default_data_space=default_data_space,
        )

    for image_spec in image_specs:
        image_spec.autodoc(output, flatten=flatten)
        logging.info("Successfully created docs for %s", image_spec.path)


@cli.command(
    name="required-packages",
    help="""Detect the Python packages required to run the
specified workflows and return them and their versions""",
)
@click.argument("task_locations", nargs=-1)
def required_packages(task_locations):
    required_modules = set()
    for task_location in task_locations:
        workflow = ClassResolver(
            TaskBase, alternative_types=[ty.Callable], package=PACKAGE_NAME
        )(task_location)
        pydra_asdict(workflow, required_modules)

    for pkg in package_from_module(required_modules):
        click.echo(f"{pkg.key}=={pkg.version}")


@cli.command(
    name="inspect-docker-exec", help="""Extract the executable from a Docker image"""
)
@click.argument("image_tag", type=str)
def inspect_docker_exec(image_tag):
    """Pulls a given Docker image tag and inspects the image to get its
    entrypoint/cmd

    IMAGE_TAG is the tag of the Docker image to inspect"""
    dc = docker.from_env()

    dc.images.pull(image_tag)

    image_attrs = dc.api.inspect_image(image_tag)["Config"]

    executable = image_attrs["Entrypoint"]
    if executable is None:
        executable = image_attrs["Cmd"]

    click.echo(executable)


@cli.command(
    help="""Displays the changelogs found in the release manifest of a deployment build

MANIFEST_JSON is a JSON file containing a list of container images built in the release
and the commands present in them"""
)
@click.argument("manifest_json", type=click.File())
@click.argument("images", nargs=-1)
def changelog(manifest_json):
    manifest = json.load(manifest_json)

    for entry in manifest["images"]:
        click.echo(
            f"{entry['name']} [{entry['version']}] changes "
            f"from {entry['previous_version']}:\n{entry['changelog']}"
        )


# @cli.command(
#     name="install-license",
#     help="""Installs a license within a store (i.e. site-wide) or dataset (project-specific)
# for use in a deployment pipeline

# LICENSE_NAME the name of the license to upload. Must match the name of the license specified
# in the deployment specification

# SOURCE_FILE path to the license file to upload

# INSTALL_LOCATIONS a list of installation locations, which are either the "nickname" of a
# store (as saved by `pydra2app store add`) or the ID of a dataset in form
# <store-nickname>//<dataset-id>[@<dataset-name>], where the dataset ID
# is either the location of the root directory (for file-system based stores) or the
# project ID for managed data repositories.
# """,
# )
# @click.argument("license_name")
# @click.argument("source_file", type=click.Path(exists=True, path_type=Path))
# @click.argument("install_locations", nargs=-1)
# @click.option(
#     "--logfile",
#     default=None,
#     type=click.Path(path_type=Path),
#     help="Log output to file instead of stdout",
# )
# @click.option("--loglevel", default="info", help="The level to display logs at")
# def install_license(install_locations, license_name, source_file, logfile, loglevel):
#     logging.basicConfig(filename=logfile, level=getattr(logging, loglevel.upper()))

#     if isinstance(source_file, bytes):  # FIXME: This shouldn't be necessary
#         source_file = Path(source_file.decode("utf-8"))

#     if not install_locations:
#         install_locations = ["dirtree"]

#     for install_loc in install_locations:
#         if "//" in install_loc:
#             dataset = Dataset.load(install_loc)
#             store_name, _, _ = Dataset.parse_id_str(install_loc)
#             msg = f"for '{dataset.name}' dataset on {store_name} store"
#         else:
#             store = DataStore.load(install_loc)
#             dataset = store.site_licenses_dataset()
#             if dataset is None:
#                 raise ValueError(
#                     f"{install_loc} store doesn't support the installation of site-wide "
#                     "licenses, please specify a dataset to install it for"
#                 )
#             msg = f"site-wide on {install_loc} store"

#         dataset.install_license(license_name, source_file)
#         logger.info("Successfully installed '%s' license %s", license_name, msg)


@cli.command(
    name="pipeline-entrypoint",
    help="""Loads/creates a dataset, then applies and launches a pipeline
in a single command. To be used within the command configuration of an XNAT
Container Service ready Docker image.

DATASET_LOCATOR string containing the nickname of the data store, the ID of the
dataset (e.g. XNAT project ID or file-system directory) and the dataset's name
in the format <store-nickname>//<dataset-id>[@<dataset-name>]

""",
)
@click.argument("dataset_locator")
@entrypoint_opts.data_columns
@entrypoint_opts.parameterisation
@entrypoint_opts.execution
@entrypoint_opts.dataset_config
@entrypoint_opts.debugging
def pipeline_entrypoint(
    dataset_locator,
    spec_path,
    **kwargs,
):
    image_spec = App.load(spec_path)

    image_spec.command.execute(
        dataset_locator,
        **kwargs,
    )


@cli.group()
def ext():
    """Command-line group for extension hooks"""


@cli.command(
    name="bootstrap",
    help="""Generate a YAML specification file for a Pydra2App App""",
)
@click.argument("output_file", type=click.Path(path_type=Path))
@click.option("--title", "-t", type=str, default=None, help="The title of the image")
@click.option(
    "--docs-url",
    "-u",
    type=str,
    default="https://place-holder.url",
    help="URL explaining the tool/workflow that is being wrapped into an app",
)
@click.option(
    "--registry",
    "-r",
    type=str,
    default="docker.io",
    help="The Docker registry of the image",
)
@click.option(
    "--description",
    "-d",
    type=str,
    default=None,
    help="A longer form description of the tool/workflow implemented in the pipeline",
)
@click.option(
    "--author",
    "-a",
    "authors",
    nargs=2,
    multiple=True,
    type=str,
    metavar="<name> <email>",
    help="The name of the author of the image",
)
@click.option(
    "--base-image",
    "-b",
    type=str,
    nargs=2,
    multiple=True,
    metavar="<attr> <value>",
    help=(
        "Set one of the attributes of the base-image, e.g. '--base-image name debian', "
        "'--base-image package_manager apt', '--base-image tag focal', "
        "'--base-image conda_env base', or '--base-image python /usr/bin/python3.7'"
    ),
)
@click.option(
    "--version", "-v", type=str, default="0.1", help="The version of the image"
)
@click.option(
    "--command-task",
    "-t",
    type=str,
    default=None,
    help="The command to execute in the image",
)
@click.option(
    "--packages-pip",
    "-y",
    type=str,
    multiple=True,
    metavar="<package-name>[==<version>]",
    help="Packages to install via pip",
)
@click.option(
    "--packages-system",
    "-s",
    type=str,
    multiple=True,
    metavar="<package-name>[==<version>]",
    help="Packages to install via the system package manager",
)
@click.option(
    "--packages-neurodocker",
    "-n",
    type=str,
    multiple=True,
    metavar="<package-name>[==<version>]",
    help="Packages to install via NeuroDocker",
)
@click.option(
    "--command-input",
    "-i",
    "command_inputs",
    type=str,
    multiple=True,
    nargs=2,
    metavar="<name> <attrs>",
    help=(
        "Input specifications, name and attribute pairs. Attributes are comma-separated "
        "name/value pairs, e.g. "
        "'datatype=str,configuration.argstr=,configuration.position=0,help=The input image''"
    ),
)
@click.option(
    "--command-output",
    "-o",
    "command_outputs",
    type=str,
    multiple=True,
    nargs=2,
    metavar="<name> <attrs>",
    help=(
        "Output specifications, name and attribute pairs. Attributes are comma-separated "
        "name/value pairs, e.g. "
        "'datatype=str,configuration.argstr=,configuration.position=1,help=The output image'"
    ),
)
@click.option(
    "--command-parameter",
    "-p",
    "command_parameters",
    type=str,
    multiple=True,
    nargs=2,
    metavar="<name> <attrs>",
    help=(
        "Parameter specifications, name and attribute pairs. Attributes are comma-separated "
        "name/value pairs, e.g. 'datatype=str,help='compression level'"
    ),
)
@click.option(
    "--command-configuration",
    "-c",
    type=str,
    multiple=True,
    nargs=2,
    metavar="<name> <value>",
    help="Command configuration value",
)
@click.option(
    "--frequency",
    "-f",
    type=str,
    default="common:Clinical[session]",
    help=(
        "The level in the data tree that the pipeline will operate on, e.g. "
        "common:Clinical[session] designates that the pipeline runs on 'sessions' "
        "as opposed to 'subjects'"
    ),
)
@click.option(
    "--license",
    "-l",
    "licenses",
    nargs=4,
    multiple=True,
    type=str,
    metavar="<license-name> <path-to-license-file> <info-url> <description>",
    help=(
        "Licenses that are required at runtime within the image. The name is used to "
        "refer to the license, when providing a license file at build time or alternatively "
        "installing the license in the data store. The path to the license file is where the "
        "license will be installed within the image. The info URL is where the details of the "
        "license can be found and where it can be acquired from. The description gives a brief "
        "description of the license and what it is required for"
    ),
)
def bootstrap(
    output_file: str,
    title: str,
    docs_url: str,
    registry: str,
    authors: ty.List[ty.Tuple[str, str]],
    base_image: ty.List[ty.Tuple[str, str]],
    version: str,
    description: str,
    command_task: str,
    packages_pip: ty.List[ty.Tuple[str, str]],
    packages_system: ty.List[ty.Tuple[str, str]],
    packages_neurodocker: ty.List[ty.Tuple[str, str]],
    command_inputs: ty.List[ty.Tuple[str, str, str]],
    command_outputs: ty.List[ty.Tuple[str, str, str]],
    command_parameters: ty.List[ty.Tuple[str, str, str]],
    command_configuration: ty.List[ty.Tuple[str, str]],
    frequency: str,
    licenses: ty.List[ty.Tuple[str, str, str, str]],
):

    # Make the output directory if it doesn't exist
    Path(output_file).parent.mkdir(parents=True, exist_ok=True)

    def unwrap_fields(fields: ty.List[ty.Tuple[str, str, str]]):
        fields_dict = {}
        for field_name, attrs_str in fields:
            attrs = [re.split(r"(?<!\\)=", a) for a in re.split(r"(?<!\\),", attrs_str)]
            unwrap_attrs = defaultdict(dict)
            for name, value in attrs:
                try:
                    value = int(value)
                except ValueError:
                    try:
                        value = float(value)
                    except ValueError:
                        pass
                if "." in name:
                    parts = name.split(".")
                    dct = unwrap_attrs[parts[0]]
                    for part in parts[1:-1]:
                        dct = dct[part]
                    dct[parts[-1]] = value
                else:
                    unwrap_attrs[name] = value
            unwrap_attrs["help"] = ""
            fields_dict[field_name] = dict(unwrap_attrs)
        return fields_dict

    ver_split_re = re.compile(r">=|<=|==|>|<")

    def split_versions(packages):
        return dict(
            ver_split_re.split(p, maxsplit=1) if "=" in p else [p, None]
            for p in packages
        )

    spec = {
        "schema_version": App.SCHEMA_VERSION,
        "title": title,
        "version": {
            "package": version,
            "build": 1,
        },
        "registry": registry,
        "docs": {
            "description": description,
            "info_url": docs_url,
        },
        "authors": [{"name": a[0], "email": a[1]} for a in authors],
        "base_image": dict(base_image),
        "packages": {
            "pip": split_versions(packages_pip),
            "system": split_versions(packages_system),
            "neurodocker": split_versions(packages_neurodocker),
        },
        "command": {
            "task": command_task,
            "row_frequency": frequency,
            "inputs": unwrap_fields(command_inputs),
            "outputs": unwrap_fields(command_outputs),
            "parameters": unwrap_fields(command_parameters),
            "configuration": dict(command_configuration),
        },
        "licenses": {
            lc[0]: {"destination": lc[1], "info_url": lc[2], "description": lc[3]}
            for lc in licenses
        },
    }

    with open(output_file, "w") as f:
        yaml.dump(spec, f)


# Ensure that all sub-packages under CLI are loaded so they are added to the
# base command
extensions = list(submodules(pydra2app, subpkg="cli"))


if __name__ == "__main__":
    make(sys.argv[1:])



from __future__ import annotations
import typing as ty
import logging
from pathlib import Path, PosixPath
import tempfile
import tarfile
import docker

logger = logging.getLogger("pydra2app")


def extract_file_from_docker_image(
    image_tag, file_path: PosixPath, out_path: ty.Optional[Path] = None
) -> Path:
    """Extracts a file from a Docker image onto the local host

    Parameters
    ----------
    image_tag : str
        the name/tag of the image to extract the file from
    file_path : PosixPath
        the path to the file inside the image

    Returns
    -------
    Path or None
        path to the extracted file or None if image doesn't exist
    """
    tmp_dir = Path(tempfile.mkdtemp())
    if out_path is None:
        out_path = tmp_dir / "extracted-dir"
    dc = docker.from_env()

    def get_image(tag):
        for img in dc.images.list():
            if tag in img.tags:
                return img
        return None

    img = get_image(image_tag)

    if img is None:
        try:
            dc.api.pull(image_tag)
        except docker.errors.APIError as e:
            if e.response.status_code == 500:
                logger.warning(
                    "Could not pull %s image from registry, attempting to check changes "
                    "with local version",
                    image_tag,
                )
            else:
                raise
    try:
        container = dc.containers.get(dc.api.create_container(image_tag)["Id"])
    except docker.errors.APIError as e:
        if e.response.status_code == 404:
            return None
        else:
            raise
    try:
        tarfile_path = tmp_dir / "tar-file.tar.gz"
        with open(tarfile_path, mode="w+b") as f:
            try:
                stream, _ = dc.api.get_archive(container.id, str(file_path))
            except docker.errors.NotFound:
                pass
            else:
                for chunk in stream:
                    f.write(chunk)
                f.flush()
    finally:
        container.remove()
    with tarfile.open(tarfile_path) as f:
        f.extractall(out_path)
    return out_path / PosixPath(file_path).name


def is_relative_to(a: Path, b: Path) -> bool:
    """Check if path a is relative to path b. Only required for Python < 3.9, after 3.8
    support is dropped can use pathlib.Path.is_relative_to instead.

    Parameters
    ----------
    a : Path
        the path to check if it is relative to b
    b : Path
        the path to check if a is relative to

    Returns
    -------
    bool
        True if a is relative to b, False otherwise
    """
    try:
        a.relative_to(b)
    except ValueError:
        return False
    else:
        return True


DOCKER_HUB = "docker.io"



from __future__ import annotations
from typing import Union, Dict, Tuple
import typing as ty
import shutil
import re
import yaml
import tempfile
from pathlib import Path
import attrs
import pytest
import docker
from pydra2app.core.image import App
from pydra2app.core.cli import (
    make,
    make_docs,
    bootstrap,
)
from frametree.core.utils import show_cli_trace
from pydra2app.core.exceptions import Pydra2AppBuildError


# @pytest.mark.xfail(reason="Need to fix a couple of things after syntax changes")
def test_deploy_make_cli(command_spec, cli_runner, work_dir):

    DOCKER_ORG = "testorg"
    DOCKER_REGISTRY = "test.registry.org"
    IMAGE_GROUP_NAME = "testpkg"

    concatenate_spec = {
        "title": "a test image spec",
        "command": command_spec,
        "version": {"package": "1.0", "build": "1"},
        "packages": {
            "system": ["vim"],  # just to test it out
            "pip": {"pydra": None},  # just to test it out
        },
        "authors": [{"name": "Some One", "email": "some.one@an.email.org"}],
        "docs": {
            "info_url": "http://concatenate.readthefakedocs.io",
        },
    }

    build_dir = work_dir / "build"
    build_dir.mkdir()
    spec_path = work_dir / DOCKER_ORG
    sub_dir = spec_path / IMAGE_GROUP_NAME
    sub_dir.mkdir(parents=True)
    with open(sub_dir / "concatenate.yml", "w") as f:
        yaml.dump(concatenate_spec, f)

    result = cli_runner(
        make,
        [
            "common:App",
            str(spec_path),
            "--build-dir",
            str(build_dir),
            "--registry",
            DOCKER_REGISTRY,
            "--loglevel",
            "warning",
            "--use-local-packages",
            "--install-extras",
            "test",
            "--raise-errors",
            "--for-localhost",
            "--dont-check-registry",
        ],
    )
    assert result.exit_code == 0, show_cli_trace(result)
    tag = result.output.strip().splitlines()[-1]
    assert tag == f"{DOCKER_REGISTRY}/{DOCKER_ORG}/{IMAGE_GROUP_NAME}.concatenate:1.0-1"

    # Clean up the built image
    dc = docker.from_env()
    dc.images.remove(tag)


# @pytest.mark.xfail(reason="Need to fix the test handle invalid docker tag name used")
def test_deploy_remake_cli(command_spec, docker_registry, cli_runner, run_prefix):
    """Tests the check to see whether"""

    IMAGE_GROUP_NAME = "testpkg-rebuild" + run_prefix

    def build_spec(spec, **kwargs):
        work_dir = Path(tempfile.mkdtemp())
        build_dir = work_dir / "build"
        build_dir.mkdir()
        spec_path = work_dir / "testorg"
        sub_dir = spec_path / IMAGE_GROUP_NAME
        sub_dir.mkdir(parents=True)
        with open(sub_dir / "concatenate.yml", "w") as f:
            yaml.dump(spec, f)

        result = cli_runner(
            make,
            [
                "common:App",
                str(spec_path),
                "--build-dir",
                str(build_dir),
                "--registry",
                docker_registry,
                "--loglevel",
                "warning",
                "--use-local-packages",
                "--install-extras",
                "test",
                "--raise-errors",
                "--check-registry",
                "--for-localhost",
            ],
            **kwargs,
        )
        return result

    concatenate_spec = {
        "title": "a test image",
        "command": command_spec,
        "version": {"package": "1.0", "build": "1"},
        "packages": {"system": ["vim"]},
        "name": "test_deploy_rebuild_cli",
        "authors": [{"name": "Some One", "email": "some.one@an.email.org"}],
        "docs": {"info_url": "http://concatenate.readthefakedocs.io"},
    }

    # Build a basic image
    result = build_spec(concatenate_spec)
    assert result.exit_code == 0, show_cli_trace(result)
    assert result.output
    tag = result.output.strip().splitlines()[-1]
    try:
        dc = docker.from_env()
        dc.api.push(tag)

        # FIXME: Need to ensure that logs are captured properly then we can test this
        # result = build_spec(concatenate_spec)
        # assert "Skipping" in result.output

        # Modify the spec so it doesn't match the original that has just been
        # built (but don't increment the version number -> image tag so there
        # is a clash)
        concatenate_spec["packages"] = {"system": ["vim", "git"]}

        with pytest.raises(Pydra2AppBuildError) as excinfo:
            build_spec(concatenate_spec, catch_exceptions=False)

        assert "doesn't match the one that was used to build the image" in str(
            excinfo.value
        )

        # Increment the build number to avoid the clash
        concatenate_spec["version"]["build"] = "2"

        result = build_spec(concatenate_spec)
        assert result.exit_code == 0, show_cli_trace(result)
        rebuilt_tag = result.output.strip().splitlines()[-1]
        dc.images.remove(rebuilt_tag)
    finally:
        # Clean up the built images
        dc.images.remove(tag)


@attrs.define
class DocsFixture:

    yaml_src: str
    markdown: str
    licenses_to_provide: ty.List[str] = attrs.field(factory=list)


docs_fixtures = {
    "simple": DocsFixture(
        """
title: a simple app
version:
  package: &package_version '0.16.1'
authors:
  - name: author_name
    email: author@email.org
base_image:
  name: abc
  tag: *package_version
  package_manager: apt
docs:
  info_url: https://example.com
  description: >-
    a test of the YAML join functionality
command:
  task: frametree.testing.tasks:identity_file
  row_frequency: common:Samples[sample]
  inputs:
    in_file:
      datatype: text/text-file
      help: the input file
  outputs:
    out_file:
      datatype: text/text-file
      help: the output file
    """.strip(),
        """
---
source_file: spec.yaml
title: package.spec
weight: 10

---

## Package Info
|Key|Value|
|---|-----|
|Name|package.spec|
|Title|a simple app|
|Package version|0.16.1|
|Build|0|
|Base image|`abc:0.16.1`|
|Maintainer|author_name (author@email.org)|
|Info URL|https://example.com|

a test of the YAML join functionality

## Command
|Key|Value|
|---|-----|
|Task|frametree.testing.tasks:identity_file|
|Operates on|sample|
#### Inputs
|Name|Required data-type|Default column data-type|Description|
|----|------------------|------------------------|-----------|
|`in_file`|<span data-toggle="tooltip" data-placement="bottom" title="text/text-file" aria-label="text/text-file">text/text-file</span>|<span data-toggle="tooltip" data-placement="bottom" title="text/text-file" aria-label="text/text-file">text/text-file</span>|the input file|

#### Outputs
|Name|Required data-type|Default column data-type|Description|
|----|------------------|------------------------|-----------|
|`out_file`|<span data-toggle="tooltip" data-placement="bottom" title="text/text-file" aria-label="text/text-file">text/text-file</span>|<span data-toggle="tooltip" data-placement="bottom" title="text/text-file" aria-label="text/text-file">text/text-file</span>|the output file|

#### Parameters
|Name|Data type|Description|
|----|---------|-----------|
""".strip(),
    ),
    "full": DocsFixture(
        """
title: a more involved image spec
version:
  package: &package_version '0.16.1'
  build: '10'
authors:
  - name: author_name
    email: author@email.org
base_image:
  name: abc
  tag: *package_version
  package_manager: yum
docs:
  info_url: https://example.com
  description: >-
    a longer description
  known_issues:
    - description: Memory overrun on large file paths
      url: https://github.com/myorg/mypackage/issues/644
packages:
  system:
    vim: 99.1
    git:
  pip:
    - pydra
    - pydra-dcm2niix
  neurodocker:
    dcm2niix: v1.0.20201102
licenses:
  freesurfer:
    destination: /opt/freesurfer/license.txt
    description: >
      license description
    info_url: http://path.to.license.provider.org/licenses
command:
    task: bids:bids_app
    inputs:
      T1w:
        configuration:
          path: anat/T1w
        datatype: medimage/nifti-gz-x
        help: "T1-weighted anatomical scan"
        column_defaults:
          datatype: medimage/dicom-series
      T2w:
        configuration:
          path: anat/T2w
        datatype: medimage/nifti-gz-x
        help: "T2-weighted anatomical scan"
        column_defaults:
          datatype: medimage/dicom-series
      fMRI:
        datatype: medimage/nifti-gz-x
        help: "functional MRI"
        configuration:
          path: func/bold/task=rest
        column_defaults:
          datatype: medimage/dicom-series
    outputs:
      mriqc:
        datatype: generic/directory
        help: "MRIQC output directory"
        configuration:
          path: mriqc
    parameters:
      fmriprep_flags:
        field: flags
        datatype: field/text
        help: description of flags param
    row_frequency: common:Clinical[session]
    configuration:
      executable: /usr/local/miniconda/bin/mriqc
      dataset: /work/bids-dataset
      app_output_dir: /work/bids-app-output
    """.strip(),
        """
---
source_file: /var/folders/mz/yn83q2fd3s758w1j75d2nnw80000gn/T/tmp47_dxmyq/specs/spec.yaml
title: package.spec
weight: 10

---

## Package Info
|Key|Value|
|---|-----|
|Name|package.spec|
|Title|a more involved image spec|
|Package version|0.16.1|
|Build|10|
|Base image|`abc:0.16.1`|
|Maintainer|author_name (author@email.org)|
|Info URL|https://example.com|
|Known issues|https://github.com/myorg/mypackage/issues/644|

a longer description

### Required licenses
|Name|URL|Description|
|----|---|-----------|
|freesurfer|`http://path.to.license.provider.org/licenses`|license description|

## Command
|Key|Value|
|---|-----|
|Task|bids:bids_app|
|Operates on|session|
#### Inputs
|Name|Required data-type|Default column data-type|Description|
|----|------------------|------------------------|-----------|
|`T1w`|<span data-toggle="tooltip" data-placement="bottom" title="medimage/nifti-gz-x" aria-label="medimage/nifti-gz-x">medimage/nifti-gz-x</span>|<span data-toggle="tooltip" data-placement="bottom" title="medimage/dicom-series" aria-label="medimage/dicom-series">medimage/dicom-series</span>|T1-weighted anatomical scan|
|`T2w`|<span data-toggle="tooltip" data-placement="bottom" title="medimage/nifti-gz-x" aria-label="medimage/nifti-gz-x">medimage/nifti-gz-x</span>|<span data-toggle="tooltip" data-placement="bottom" title="medimage/dicom-series" aria-label="medimage/dicom-series">medimage/dicom-series</span>|T2-weighted anatomical scan|
|`fMRI`|<span data-toggle="tooltip" data-placement="bottom" title="medimage/nifti-gz-x" aria-label="medimage/nifti-gz-x">medimage/nifti-gz-x</span>|<span data-toggle="tooltip" data-placement="bottom" title="medimage/dicom-series" aria-label="medimage/dicom-series">medimage/dicom-series</span>|functional MRI|

#### Outputs
|Name|Required data-type|Default column data-type|Description|
|----|------------------|------------------------|-----------|
|`mriqc`|<span data-toggle="tooltip" data-placement="bottom" title="generic/directory" aria-label="generic/directory">generic/directory</span>|<span data-toggle="tooltip" data-placement="bottom" title="generic/directory" aria-label="generic/directory">generic/directory</span>|MRIQC output directory|

#### Parameters
|Name|Data type|Description|
|----|---------|-----------|
|`fmriprep_flags`|`str`|description of flags param|
""".strip(),
        ["freesurfer"],
    ),
}


@pytest.mark.parametrize("fixture", docs_fixtures.items(), ids=lambda x: x[0])
def test_make_docs_cli(
    cli_runner, run_prefix, work_dir: Path, fixture: Tuple[str, DocsFixture]
):
    fixture_name, fixture_content = fixture

    # TODO handle multiple 'files' in a fixture
    print(f"Processing fixture: {fixture_name!r}")
    output = _make_docs(cli_runner, work_dir, fixture_content.yaml_src)

    strip_source_file_re = re.compile(r"source_file:.*")

    stripped_output = strip_source_file_re.sub("", output)
    stripped_reference = strip_source_file_re.sub("", fixture_content.markdown)

    assert (
        stripped_output == stripped_reference
    ), f"Fixture {fixture_name!r} didn't match output"


def _make_docs(
    cli_runner,
    work_dir: Path,
    docs: Union[str, Dict[str, str]],
    *args,
    flatten: ty.Optional[bool] = None,
) -> Union[str, Dict[str, str]]:
    out_dir = work_dir / "out"
    org_dir = work_dir / "org"
    specs_dir = org_dir / "package"
    if specs_dir.exists():
        shutil.rmtree(specs_dir)
    specs_dir.mkdir(parents=True)

    if type(docs) is str:
        (specs_dir / "spec.yaml").write_text(docs)
    else:
        for name, content in docs.items():
            path = specs_dir / name
            path.parent.mkdir(parents=True, exist_ok=True)
            path.write_text(content)

    result = cli_runner(
        make_docs,
        [
            specs_dir.as_posix(),
            out_dir.as_posix(),
            "--spec-root",
            str(work_dir),
        ]
        + (["--flatten" if flatten else "--no-flatten"] if flatten is not None else [])
        + list(args),
    )

    assert result.exit_code == 0, show_cli_trace(result)

    if type(docs) is str:
        return (out_dir / "org" / "package.spec.md").read_text().strip()
    else:
        return {
            file.relative_to(out_dir).as_posix(): file.read_text().strip()
            for file in out_dir.glob("*.md")
        }


def test_bootstrap(cli_runner, work_dir):

    out_yaml = work_dir / "out-spec.yaml"

    result = cli_runner(
        bootstrap,
        [
            str(out_yaml),
            "--author",
            "An Author",
            "$an.author@a.institution",
            "--version",
            "7.1.1",
            "--base-image",
            "name",
            "vnmd/freesurfer_7.1.1",
            "--base-image",
            "tag",
            "20211216",
            "--base-image",
            "package_manager",
            "apt",
            "--packages-pip",
            "fileformats-medimage-extras==0.8.0",
            "--packages-neurodocker",
            "dcm2niix==v1.0.20201102",
            "--command-task",
            "common:shell",
            "--command-input",
            "head",
            "datatype=medimage/nifti-gz,configuration.position=-2,configuration.argstr=''",
            "--command-output",
            "brain",
            "datatype=medimage/nifti-gz,configuration.position=-1,configuration.argstr=''",
            "--command-configuration",
            "executable",
            "mri_convert",
            "--title",
            "MRI Convert",
            "--license",
            "freesurfer",
            "/opt/freesurfer/license.txt",
            "https://surfer.nmr.mgh.harvard.edu/registration.html",
            "The freesurfer license that is required to run any tool within the package",
        ],
    )

    assert result.exit_code == 0, show_cli_trace(result)

    # Check that the YAML loads properly
    App.load(out_yaml)
